/*
 * Copyright (C) 2024 iopsys Software Solutions AB
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2.1
 * as published by the Free Software Foundation
 *
 *	Author: Suvendhu Hansa <suvendhu.hansa@iopsys.eu>
 */

#include "parentalcontrol.h"

char *AccessPolicy[] = {"Allow", "Deny", NULL};
char *Days[] = {"Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday", NULL};

/*************************************************************
* ENTRY METHOD
**************************************************************/
static int browseParentalControlURLBundleInst(struct dmctx *dmctx, DMNODE *parent_node, void *prev_data, char *prev_instance)
{
	struct dmmap_dup *p = NULL;
	char *inst = NULL;
	LIST_HEAD(dup_list);

	synchronize_specific_config_sections_with_dmmap("parentalcontrol", "urlbundle", "dmmap_parentalcontrol", &dup_list);
	list_for_each_entry(p, &dup_list, list) {

		inst = handle_instance(dmctx, parent_node, p->dmmap_section, "instance", "alias");

		if (DM_LINK_INST_OBJ(dmctx, parent_node, (void *)p, inst) == DM_STOP)
			break;
	}
	free_dmmap_config_dup_list(&dup_list);
	return 0;
}

static int browseParentalControlProfileInst(struct dmctx *dmctx, DMNODE *parent_node, void *prev_data, char *prev_instance)
{
	struct dmmap_dup *p = NULL;
	char *inst = NULL;
	LIST_HEAD(dup_list);

	synchronize_specific_config_sections_with_dmmap("parentalcontrol", "profile", "dmmap_parentalcontrol", &dup_list);
	list_for_each_entry(p, &dup_list, list) {

		inst = handle_instance(dmctx, parent_node, p->dmmap_section, "instance", "alias");

		if (DM_LINK_INST_OBJ(dmctx, parent_node, (void *)p, inst) == DM_STOP)
			break;
	}
	free_dmmap_config_dup_list(&dup_list);
	return 0;
}

static int browseParentalControlProfileURLFilterInst(struct dmctx *dmctx, DMNODE *parent_node, void *prev_data, char *prev_instance)
{
	struct dmmap_dup *p = NULL;
	char *inst = NULL;
	LIST_HEAD(dup_list);

	synchronize_specific_config_sections_with_dmmap("parentalcontrol", "profile_urlfilter", "dmmap_parentalcontrol", &dup_list);
	list_for_each_entry(p, &dup_list, list) {
		char *parent = NULL, *cur_sec = NULL;

		dmuci_get_value_by_section_string(p->config_section, "dm_parent", &parent);
		cur_sec = section_name(((struct dmmap_dup *)prev_data)->config_section);

		if (DM_STRCMP(cur_sec, parent) != 0)
			continue;
		
		inst = handle_instance(dmctx, parent_node, p->dmmap_section, "instance", "alias");

		if (DM_LINK_INST_OBJ(dmctx, parent_node, (void *)p, inst) == DM_STOP)
			break;
	}
	free_dmmap_config_dup_list(&dup_list);
	return 0;
}

static int browseParentalControlProfileBedTimeScheduleInst(struct dmctx *dmctx, DMNODE *parent_node, void *prev_data, char *prev_instance)
{
	struct dmmap_dup *p = NULL;
	char *inst = NULL;
	LIST_HEAD(dup_list);

	synchronize_specific_config_sections_with_dmmap("parentalcontrol", "profile_bedtime_schedule", "dmmap_parentalcontrol", &dup_list);
	list_for_each_entry(p, &dup_list, list) {
		char *parent = NULL, *cur_sec = NULL;

		dmuci_get_value_by_section_string(p->config_section, "dm_parent", &parent);
		cur_sec = section_name(((struct dmmap_dup *)prev_data)->config_section);

		if (DM_STRCMP(cur_sec, parent) != 0)
			continue;
		
		inst = handle_instance(dmctx, parent_node, p->dmmap_section, "instance", "alias");

		if (DM_LINK_INST_OBJ(dmctx, parent_node, (void *)p, inst) == DM_STOP)
			break;
	}
	free_dmmap_config_dup_list(&dup_list);
	return 0;
}

/*************************************************************
* ADD & DEL OBJ
**************************************************************/
static int addObjParentalControlURLBundle(char *refparam, struct dmctx *ctx, void *data, char **instance)
{
	struct uci_section *s = NULL, *dmmap_s = NULL;
	char s_name[64];

	snprintf(s_name, sizeof(s_name), "urlbundle_%s", *instance);

	dmuci_add_section("parentalcontrol", "urlbundle", &s);
	dmuci_rename_section_by_section(s, s_name);

	dmuci_add_section_bbfdm("dmmap_parentalcontrol", "urlbundle", &dmmap_s);
	dmuci_set_value_by_section(dmmap_s, "section_name", section_name(s));
	dmuci_set_value_by_section(dmmap_s, "instance", *instance);
	return 0;
}

static int delObjParentalControlURLBundle(char *refparam, struct dmctx *ctx, void *data, char *instance, unsigned char del_action)
{
	struct uci_section *s = NULL, *stmp = NULL;

	switch (del_action) {
		case DEL_INST:
			dmuci_delete_by_section(((struct dmmap_dup *)data)->config_section, NULL, NULL);
			dmuci_delete_by_section(((struct dmmap_dup *)data)->dmmap_section, NULL, NULL);
			break;
		case DEL_ALL:
			uci_foreach_sections_safe("parentalcontrol", "urlbundle", stmp, s) {
				struct uci_section *dmmap_section = NULL;

				get_dmmap_section_of_config_section("dmmap_parentalcontrol", "urlbundle", section_name(s), &dmmap_section);
				dmuci_delete_by_section(dmmap_section, NULL, NULL);

				dmuci_delete_by_section(s, NULL, NULL);
			}
			break;
	}
	return 0;
}

static int addObjParentalControlProfile(char *refparam, struct dmctx *ctx, void *data, char **instance)
{
	struct uci_section *s = NULL, *dmmap_s = NULL;
	char s_name[64] = {0};

	snprintf(s_name, sizeof(s_name), "profile_%s", *instance);

	dmuci_add_section("parentalcontrol", "profile", &s);
	dmuci_rename_section_by_section(s, s_name);

	dmuci_add_section_bbfdm("dmmap_parentalcontrol", "profile", &dmmap_s);
	dmuci_set_value_by_section(dmmap_s, "section_name", s_name);
	dmuci_set_value_by_section(dmmap_s, "instance", *instance);
	return 0;
}

static int delObjParentalControlProfile(char *refparam, struct dmctx *ctx, void *data, char *instance, unsigned char del_action)
{
	struct uci_section *s = NULL, *stmp = NULL;

	switch (del_action) {
		case DEL_INST:
			dmuci_delete_by_section(((struct dmmap_dup *)data)->config_section, NULL, NULL);
			dmuci_delete_by_section(((struct dmmap_dup *)data)->dmmap_section, NULL, NULL);
			break;
		case DEL_ALL:
			uci_foreach_sections_safe("parentalcontrol", "profile", stmp, s) {
				struct uci_section *dmmap_section = NULL;

				get_dmmap_section_of_config_section("dmmap_parentalcontrol", "profile", section_name(s), &dmmap_section);
				dmuci_delete_by_section(dmmap_section, NULL, NULL);

				dmuci_delete_by_section(s, NULL, NULL);
			}
			break;
	}
	return 0;
}

static int addObjParentalControlProfileURLFilter(char *refparam, struct dmctx *ctx, void *data, char **instance)
{
	struct uci_section *s = NULL, *dmmap_s = NULL;
	char s_name[64] = {0};
	char *par_inst = NULL;

	dmuci_get_value_by_section_string(((struct dmmap_dup *)data)->dmmap_section, "instance", &par_inst);
	if (DM_STRLEN(par_inst) == 0)
		return 0;

	snprintf(s_name, sizeof(s_name), "profile_%s_profile_urlfilter_%s", par_inst, *instance);

	dmuci_add_section("parentalcontrol", "profile_urlfilter", &s);
	dmuci_rename_section_by_section(s, s_name);
	dmuci_set_value_by_section(s, "dm_parent", section_name(((struct dmmap_dup *)data)->config_section));
	dmuci_set_value_by_section(s, "access", "Deny");

	dmuci_add_section_bbfdm("dmmap_parentalcontrol", "profile_urlfilter", &dmmap_s);
	dmuci_set_value_by_section(dmmap_s, "section_name", s_name);
	dmuci_set_value_by_section(dmmap_s, "dm_parent", section_name(((struct dmmap_dup *)data)->config_section));
	dmuci_set_value_by_section(dmmap_s, "instance", *instance);
	return 0;
}

static int delObjParentalControlProfileURLFilter(char *refparam, struct dmctx *ctx, void *data, char *instance, unsigned char del_action)
{
	struct uci_section *s = NULL, *stmp = NULL;

	switch (del_action) {
		case DEL_INST:
			dmuci_delete_by_section(((struct dmmap_dup *)data)->config_section, NULL, NULL);
			dmuci_delete_by_section(((struct dmmap_dup *)data)->dmmap_section, NULL, NULL);
			break;
		case DEL_ALL:
			uci_foreach_sections_safe("parentalcontrol", "profile_urlfilter", stmp, s) {
				char *parent = NULL, *cur_sec = NULL;

				dmuci_get_value_by_section_string(s, "dm_parent", &parent);
				cur_sec = section_name(((struct dmmap_dup *)data)->config_section);

				if (DM_STRCMP(cur_sec, parent) != 0)
					continue;

				struct uci_section *dmmap_section = NULL;

				get_dmmap_section_of_config_section("dmmap_parentalcontrol", "profile_urlfilter", section_name(s), &dmmap_section);
				dmuci_delete_by_section(dmmap_section, NULL, NULL);
				dmuci_delete_by_section(s, NULL, NULL);
			}
			break;
	}
	return 0;
}

static int addObjParentalControlProfileBedTimeSchedule(char *refparam, struct dmctx *ctx, void *data, char **instance)
{
	struct uci_section *s = NULL, *dmmap_s = NULL;
	char s_name[64] = {0};
	char *par_inst = NULL;

	dmuci_get_value_by_section_string(((struct dmmap_dup *)data)->dmmap_section, "instance", &par_inst);
	if (DM_STRLEN(par_inst) == 0)
		return 0;

	snprintf(s_name, sizeof(s_name), "profile_%s_profile_bedtime_schedule_%s", par_inst, *instance);

	dmuci_add_section("parentalcontrol", "profile_bedtime_schedule", &s);
	dmuci_rename_section_by_section(s, s_name);
	dmuci_set_value_by_section(s, "dm_parent", section_name(((struct dmmap_dup *)data)->config_section));

	dmuci_add_section_bbfdm("dmmap_parentalcontrol", "profile_bedtime_schedule", &dmmap_s);
	dmuci_set_value_by_section(dmmap_s, "section_name", s_name);
	dmuci_set_value_by_section(dmmap_s, "dm_parent", section_name(((struct dmmap_dup *)data)->config_section));
	dmuci_set_value_by_section(dmmap_s, "instance", *instance);
	return 0;
}

static int delObjParentalControlProfileBedTimeSchedule(char *refparam, struct dmctx *ctx, void *data, char *instance, unsigned char del_action)
{
	struct uci_section *s = NULL, *stmp = NULL;

	switch (del_action) {
		case DEL_INST:
			dmuci_delete_by_section(((struct dmmap_dup *)data)->config_section, NULL, NULL);
			dmuci_delete_by_section(((struct dmmap_dup *)data)->dmmap_section, NULL, NULL);
			break;
		case DEL_ALL:
			uci_foreach_sections_safe("parentalcontrol", "profile_bedtime_schedule", stmp, s) {
				char *parent = NULL, *cur_sec = NULL;

				dmuci_get_value_by_section_string(s, "dm_parent", &parent);
				cur_sec = section_name(((struct dmmap_dup *)data)->config_section);

				if (DM_STRCMP(cur_sec, parent) != 0)
					continue;

				struct uci_section *dmmap_section = NULL;

				get_dmmap_section_of_config_section("dmmap_parentalcontrol", "profile_bedtime_schedule", section_name(s), &dmmap_section);
				dmuci_delete_by_section(dmmap_section, NULL, NULL);
				dmuci_delete_by_section(s, NULL, NULL);
			}
			break;
	}
	return 0;
}

/*************************************************************
* GET & SET PARAM
**************************************************************/
static int get_ParentalControl_Enable(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	char *v;

	dmuci_get_option_value_string("parentalcontrol", "globals", "enable", &v);
	*value = (*v == '0') ? "0" : "1";
	return 0;
}

static int set_ParentalControl_Enable(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	bool b = true;

	switch (action) {
		case VALUECHECK:
			if (bbfdm_validate_boolean(ctx, value))
				return FAULT_9007;
			break;
		case VALUESET:
			string_to_bool(value, &b);

			dmuci_set_value("parentalcontrol", "globals", "enable", b ? "1": "0");
			break;
	}

	return 0;
}

static int get_ParentalControl_ProfileNumberOfEntries(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	int cnt = get_number_of_entries(ctx, data, instance, browseParentalControlProfileInst);
	dmasprintf(value, "%d", cnt);
	return 0;
}

static int get_ParentalControl_URLBundleNumberOfEntries(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	int cnt = get_number_of_entries(ctx, data, instance, browseParentalControlURLBundleInst);
	dmasprintf(value, "%d", cnt);
	return 0;
}

static int get_ParentalControlURLBundle_Name(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	dmuci_get_value_by_section_string(((struct dmmap_dup *)data)->config_section, "name", value);
	return 0;
}

static int set_ParentalControlURLBundle_Name(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	switch (action)	{
		case VALUECHECK:
			if (bbfdm_validate_string(ctx, value, -1, 32, NULL, NULL))
				return FAULT_9007;
			break;
		case VALUESET:
			dmuci_set_value_by_section(((struct dmmap_dup *)data)->config_section, "name", value);
			break;
	}
	return 0;
}

static int get_ParentalControlURLBundle_DownloadURL(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	dmuci_get_value_by_section_string(((struct dmmap_dup *)data)->config_section, "download_url", value);
	return 0;
}

static int set_ParentalControlURLBundle_DownloadURL(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	switch (action)	{
		case VALUECHECK:
			if (bbfdm_validate_string(ctx, value, -1, -1, NULL, NULL))
				return FAULT_9007;
			break;
		case VALUESET:
			dmuci_set_value_by_section(((struct dmmap_dup *)data)->config_section, "download_url", value);
			break;
	}
	return 0;
}

static int get_ParentalControlURLBundle_CustomURL(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	struct uci_list *val = NULL;
	dmuci_get_value_by_section_list(((struct dmmap_dup *)data)->config_section, "custom_url", &val);
	*value = dmuci_list_to_string(val, ",");
	return 0;
}

static int set_ParentalControlURLBundle_CustomURL(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	size_t length, i;
	char **arr;

	switch (action)	{
		case VALUECHECK:
			if (bbfdm_validate_string_list(ctx, value, -1, -1, -1, -1, -1, NULL, NULL))
				return FAULT_9007;
			break;
		case VALUESET:
			arr = strsplit(value, ",", &length);
			dmuci_set_value_by_section(((struct dmmap_dup *)data)->config_section, "custom_url", "");
			for (i = 0; i < length; i++)
				dmuci_add_list_value_by_section(((struct dmmap_dup *)data)->config_section, "custom_url", arr[i]);
			break;
	}
	return 0;
}

static int get_ParentalControlProfile_Alias(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	return bbf_get_alias(ctx, ((struct dmmap_dup *)data)->dmmap_section, "alias", instance, value);
}

static int set_ParentalControlProfile_Alias(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	return bbf_set_alias(ctx, ((struct dmmap_dup *)data)->dmmap_section, "alias", instance, value);
}

static int get_ParentalControlProfile_Name(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	dmuci_get_value_by_section_string(((struct dmmap_dup *)data)->config_section, "name", value);
	return 0;
}

static int set_ParentalControlProfile_Name(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	switch (action)	{
		case VALUECHECK:
			if (bbfdm_validate_string(ctx, value, -1, 64, NULL, NULL))
				return FAULT_9007;
			break;
		case VALUESET:
			dmuci_set_value_by_section(((struct dmmap_dup *)data)->config_section, "name", value);
			break;
	}
	return 0;
}

static int get_ParentalControlProfile_IconName(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	dmuci_get_value_by_section_string(((struct dmmap_dup *)data)->config_section, "icon_name", value);
	return 0;
}

static int set_ParentalControlProfile_IconName(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	switch (action)	{
		case VALUECHECK:
			if (bbfdm_validate_string(ctx, value, -1, 32, NULL, NULL))
				return FAULT_9007;
			break;
		case VALUESET:
			dmuci_set_value_by_section(((struct dmmap_dup *)data)->config_section, "icon_name", value);
			break;
	}
	return 0;
}

static int get_ParentalControlProfile_HostList(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	struct uci_list *val = NULL;
	dmuci_get_value_by_section_list(((struct dmmap_dup *)data)->config_section, "host", &val);
	*value = dmuci_list_to_string(val, ",");
	return 0;
}

static int set_ParentalControlProfile_HostList(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	size_t length, i;
	char **arr;

	switch (action)	{
		case VALUECHECK:
			if (bbfdm_validate_string_list(ctx, value, -1, -1, -1, -1, -1, NULL, NULL))
				return FAULT_9007;
			break;
		case VALUESET:
			arr = strsplit(value, ",", &length);
			dmuci_set_value_by_section(((struct dmmap_dup *)data)->config_section, "host", "");
			for (i = 0; i < length; i++)
				dmuci_add_list_value_by_section(((struct dmmap_dup *)data)->config_section, "host", arr[i]);
			break;
	}
	return 0;
}

static int get_ParentalControlProfile_URLFilterNumberOfEntries(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	int cnt = get_number_of_entries(ctx, data, instance, browseParentalControlProfileURLFilterInst);
	dmasprintf(value, "%d", cnt);
	return 0;
}

static int get_ParentalControlProfileURLFilter_Enable(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmuci_get_value_by_section_fallback_def(((struct dm_data *)data)->config_section, "enable", "0");
	return 0;
}

static int set_ParentalControlProfileURLFilter_Enable(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	bool b;

	switch (action)	{
		case VALUECHECK:
			if (bbfdm_validate_boolean(ctx, value))
				return FAULT_9007;
			break;
		case VALUESET:
			string_to_bool(value, &b);
			dmuci_set_value_by_section(((struct dm_data *)data)->config_section, "enable", b ? "1" : "0");
			break;
	}
	return 0;
}

static int get_ParentalControlProfileURLFilter_AccessPolicy(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmuci_get_value_by_section_fallback_def(((struct dm_data *)data)->config_section, "access", "Deny");
	return 0;
}

static int set_ParentalControlProfileURLFilter_AccessPolicy(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	switch (action)	{
		case VALUECHECK:
			if (bbfdm_validate_string(ctx, value, -1, -1, AccessPolicy, NULL))
				return FAULT_9007;
			break;
		case VALUESET:
			dmuci_set_value_by_section(((struct dm_data *)data)->config_section, "access", value);
			break;
	}
	return 0;
}

static int get_ParentalControlProfileURLFilter_Order(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	dmuci_get_value_by_section_string(((struct dmmap_dup *)data)->config_section, "order", value);
	return 0;
}

static int set_ParentalControlProfileURLFilter_Order(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	switch (action)	{
		case VALUECHECK:
			if (bbfdm_validate_unsignedInt(ctx, value, RANGE_ARGS{{NULL,NULL}}, 1))
				return FAULT_9007;
			break;
		case VALUESET:
			dmuci_set_value_by_section(((struct dm_data *)data)->config_section, "order", value);
			break;
	}
	return 0;
}

static int get_ParentalControlProfileURLFilter_FilterText(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	struct uci_list *val = NULL;
	dmuci_get_value_by_section_list(((struct dmmap_dup *)data)->config_section, "filter_text", &val);
	*value = dmuci_list_to_string(val, ",");
	return 0;
}

static int set_ParentalControlProfileURLFilter_FilterText(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	size_t length, i;
	char **arr;

	switch (action)	{
		case VALUECHECK:
			if (bbfdm_validate_string_list(ctx, value, -1, -1, -1, -1, -1, NULL, NULL))
				return FAULT_9007;
			break;
		case VALUESET:
			arr = strsplit(value, ",", &length);
			dmuci_set_value_by_section(((struct dmmap_dup *)data)->config_section, "filter_text", "");
			for (i = 0; i < length; i++)
				dmuci_add_list_value_by_section(((struct dmmap_dup *)data)->config_section, "filter_text", arr[i]);
			break;
	}
	return 0;
}

static int get_ParentalControlProfileURLFilter_URLBundleRef(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	char *bundle = NULL;

	dmuci_get_value_by_section_string(((struct dmmap_dup *)data)->config_section, "profile_urlbundle", &bundle);
	adm_entry_get_reference_param(ctx, "Device."BBF_VENDOR_PREFIX"ParentalControl.URLBundle.*.Name", bundle, value);

	return 0;
}

static int set_ParentalControlProfileURLFilter_URLBundleRef(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	char bundle_ref[] = "Device."BBF_VENDOR_PREFIX"ParentalControl.URLBundle.";
	char *allowed_objects[] = {
			bundle_ref,
			NULL
	};

	struct dm_reference reference = {0};
	bbf_get_reference_args(value, &reference);

	switch (action)	{
		case VALUECHECK:
			if (bbfdm_validate_string(ctx, reference.path, -1, -1, NULL, NULL))
				return FAULT_9007;

			if (dm_validate_allowed_objects(ctx, &reference, allowed_objects))
				return FAULT_9007;

			break;
		case VALUESET:
			dmuci_set_value_by_section(((struct dmmap_dup *)data)->config_section, "profile_urlbundle", reference.value);
			break;
	}
	return 0;
}

static int get_ParentalControlProfileURLFilter_ScheduleRef(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	struct uci_list *val = NULL;
	dmuci_get_value_by_section_list(((struct dmmap_dup *)data)->config_section, "profile_urlfilter_schedule", &val);

	if (val) {
		struct uci_element *e = NULL;
		char lbuf[2048] = {0};
		unsigned pos = 0;

		lbuf[0] = 0;
		uci_foreach_element(val, e) {
			if (e->name) {
				adm_entry_get_reference_param(ctx, "Device.Schedules.Schedule.*.Alias", e->name, value);
				if ((DM_STRLEN(*value) > 0) && (DM_STRLEN(*value) < (2048 - strlen(lbuf))))
					pos += (unsigned)snprintf(&lbuf[pos], sizeof(lbuf) - pos, "%s,", *value);
			}
		}

		if (pos)
			lbuf[pos - 1] = 0;

		*value = dmstrdup(lbuf);
	}

	return 0;
}

static int add_schedule_reference(struct dmctx *ctx, void *data, char *value, int action, const char *param_name)
{
	size_t length, i;
	char **arr;
	struct dm_reference reference = {0};
	char *allowed_obj[] = {
		"Device.Schedules.Schedule.",
		NULL
	};

	switch (action)	{
		case VALUECHECK:
			if (DM_STRLEN(value) == 0)
				break;

			if (bbfdm_validate_string_list(ctx, value, -1, -1, -1, -1, -1, NULL, NULL))
				return FAULT_9007;

			arr = strsplit(value, ",", &length);
			for (i = 0; i < length; i++) {
				memset(&reference, 0 , sizeof(struct dm_reference));
				bbf_get_reference_args(arr[i], &reference);

				if (DM_STRLEN(reference.path) == 0)
					continue;

				if (dm_validate_allowed_objects(ctx, &reference, allowed_obj))
					return FAULT_9007;
			}

			break;
		case VALUESET:
			dmuci_set_value_by_section(((struct dmmap_dup *)data)->config_section, (char *)param_name, "");
			if (DM_STRLEN(value) == 0)
				break;

			arr = strsplit(value, ",", &length);
			for (i = 0; i < length; i++) {
				memset(&reference, 0 , sizeof(struct dm_reference));
				bbf_get_reference_args(arr[i], &reference);

				if (DM_STRLEN(reference.path) == 0 || DM_STRLEN(reference.value) == 0)
					continue;

				dmuci_add_list_value_by_section(((struct dmmap_dup *)data)->config_section, (char *)param_name, reference.value);
			}

			break;
	}

	return 0;
}

static int set_ParentalControlProfileURLFilter_ScheduleRef(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	return add_schedule_reference(ctx, data, value, action, "profile_urlfilter_schedule");
}

static int get_ParentalControlProfileBedTime_Enable(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmuci_get_value_by_section_fallback_def(((struct dm_data *)data)->config_section, "bedtime_enable", "0");
	return 0;
}

static int set_ParentalControlProfileBedTime_Enable(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	bool b;

	switch (action)	{
		case VALUECHECK:
			if (bbfdm_validate_boolean(ctx, value))
				return FAULT_9007;
			break;
		case VALUESET:
			string_to_bool(value, &b);
			dmuci_set_value_by_section(((struct dm_data *)data)->config_section, "bedtime_enable", b ? "1" : "0");
			break;
	}
	return 0;
}

static int get_ParentalControlProfileBedTime_ScheduleNumberOfEntries(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	int cnt = get_number_of_entries(ctx, data, instance, browseParentalControlProfileBedTimeScheduleInst);
	dmasprintf(value, "%d", cnt);
	return 0;
}

static int get_ParentalControlProfileBedTimeSchedule_Alias(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	return bbf_get_alias(ctx, ((struct dmmap_dup *)data)->dmmap_section, "alias", instance, value);
}

static int set_ParentalControlProfileBedTimeSchedule_Alias(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	return bbf_set_alias(ctx, ((struct dmmap_dup *)data)->dmmap_section, "alias", instance, value);
}

static int get_ParentalControlProfileBedTimeSchedule_Enable(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	dmuci_get_value_by_section_string(((struct dmmap_dup *)data)->config_section, "enable", value);
	return 0;
}

static int set_ParentalControlProfileBedTimeSchedule_Enable(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	switch (action)	{
		case VALUECHECK:
			if (bbfdm_validate_string(ctx, value, -1, -1, NULL, NULL))
				return FAULT_9007;
			break;
		case VALUESET:
			dmuci_set_value_by_section(((struct dmmap_dup *)data)->config_section, "enable", value);
			break;
	}
	return 0;
}

static int get_ParentalControlProfileBedTimeSchedule_Name(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	dmuci_get_value_by_section_string(((struct dmmap_dup *)data)->config_section, "name", value);
	return 0;
}

static int set_ParentalControlProfileBedTimeSchedule_Name(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	switch (action)	{
		case VALUECHECK:
			if (bbfdm_validate_string(ctx, value, -1, -1, NULL, NULL))
				return FAULT_9007;
			break;
		case VALUESET:
			dmuci_set_value_by_section(((struct dmmap_dup *)data)->config_section, "name", value);
			break;
	}
	return 0;
}

static int get_ParentalControlProfileBedTimeSchedule_Days(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	struct uci_list *val = NULL;
	dmuci_get_value_by_section_list(((struct dmmap_dup *)data)->config_section, "days", &val);
	*value = dmuci_list_to_string(val, ",");
	return 0;
}

static int set_ParentalControlProfileBedTimeSchedule_Days(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	size_t length, i;
	char **arr;

	switch (action)	{
		case VALUECHECK:
			if (bbfdm_validate_string_list(ctx, value, -1, -1, -1, -1, -1, Days, NULL))
				return FAULT_9007;
			break;
		case VALUESET:
			arr = strsplit(value, ",", &length);
			dmuci_set_value_by_section(((struct dmmap_dup *)data)->config_section, "day", "");
			for (i = 0; i < length; i++)
				dmuci_add_list_value_by_section(((struct dmmap_dup *)data)->config_section, "day", arr[i]);
			break;
	}
	return 0;
}

static int get_ParentalControlProfileBedTimeSchedule_StartTime(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	dmuci_get_value_by_section_string(((struct dmmap_dup *)data)->config_section, "start_time", value);
	return 0;
}

static int set_ParentalControlProfileBedTimeSchedule_StartTime(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	switch (action)	{
		case VALUECHECK:
			if (bbfdm_validate_string(ctx, value, -1, 5, NULL, NULL))
				return FAULT_9007;

			char *reg_exp = "^([01][0-9]|2[0-3]):[0-5][0-9]$";
			if (match(value, reg_exp, 0, NULL) != true)
				return FAULT_9007;
			break;
		case VALUESET:
			dmuci_set_value_by_section(((struct dmmap_dup *)data)->config_section, "start_time", value);
			break;
	}
	return 0;
}

static int get_ParentalControlProfileBedTimeSchedule_EndTime(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	dmuci_get_value_by_section_string(((struct dmmap_dup *)data)->config_section, "end_time", value);
	return 0;
}

static int set_ParentalControlProfileBedTimeSchedule_EndTime(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	switch (action)	{
		case VALUECHECK:
			if (bbfdm_validate_string(ctx, value, -1, 5, NULL, NULL))
				return FAULT_9007;

			char *reg_exp = "^([01][0-9]|2[0-3]):[0-5][0-9]$";
			if (match(value, reg_exp, 0, NULL) != true)
				return FAULT_9007;
			break;
		case VALUESET:
			dmuci_set_value_by_section(((struct dmmap_dup *)data)->config_section, "end_time", value);
			break;
	}
	return 0;
}

static int get_ParentalControlProfileInternetBreak_Enable(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = dmuci_get_value_by_section_fallback_def(((struct dm_data *)data)->config_section, "internet_access_enable", "0");
	return 0;
}

static int set_ParentalControlProfileInternetBreak_Enable(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	bool b;

	switch (action)	{
		case VALUECHECK:
			if (bbfdm_validate_boolean(ctx, value))
				return FAULT_9007;
			break;
		case VALUESET:
			string_to_bool(value, &b);
			dmuci_set_value_by_section(((struct dm_data *)data)->config_section, "internet_access_enable", b ? "1" : "0");
			break;
	}
	return 0;
}

static int get_ParentalControlProfileInternetBreak_ScheduleRef(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	struct uci_list *val = NULL;
	dmuci_get_value_by_section_list(((struct dmmap_dup *)data)->config_section, "internet_access_schedule", &val);

	if (val) {
		struct uci_element *e = NULL;
		char lbuf[1024] = {0};
		unsigned pos = 0;

		lbuf[0] = 0;
		uci_foreach_element(val, e) {
			if (e->name) {
				adm_entry_get_reference_param(ctx, "Device.Schedules.Schedule.*.Alias", e->name, value);
				if ((DM_STRLEN(*value) > 0) && (DM_STRLEN(*value) < (1024 - strlen(lbuf))))
					pos += (unsigned)snprintf(&lbuf[pos], sizeof(lbuf) - pos, "%s,", *value);
			}
		}

		if (pos)
			lbuf[pos - 1] = 0;

		*value = dmstrdup(lbuf);
	}

	return 0;
}

static int set_ParentalControlProfileInternetBreak_ScheduleRef(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	return add_schedule_reference(ctx, data, value, action, "internet_access_schedule");
}

/**********************************************************************************************************************************
*                                            OBJ & PARAM DEFINITION
***********************************************************************************************************************************/
DM_MAP_OBJ tDynamicObj[] = {
/* parentObj, nextObject, parmeter */
{"Device.", tExtsObj, NULL},
{0}
};

DMOBJ tExtsObj[] = {
/* OBJ, permission, addobj, delobj, checkdep, browseinstobj, nextdynamicobj, dynamicleaf, nextobj, leaf, linker, bbfdm_type, uniqueKeys*/
{BBF_VENDOR_PREFIX"ParentalControl", &DMREAD, NULL, NULL, NULL, NULL, NULL, NULL, tParentalControlObj, tParentalControlParams, NULL, BBFDM_BOTH, NULL},
{0}
};

/* *** Device.{BBF_VENDOR_PREFIX}ParentalControl. *** */
DMOBJ tParentalControlObj[] = {
/* OBJ, permission, addobj, delobj, checkdep, browseinstobj, nextdynamicobj, dynamicleaf, nextobj, leaf, linker, bbfdm_type, uniqueKeys */
{"URLBundle", &DMWRITE, addObjParentalControlURLBundle, delObjParentalControlURLBundle, NULL, browseParentalControlURLBundleInst, NULL, NULL, NULL, tParentalControlURLBundleParams, NULL, BBFDM_BOTH, NULL},
{"Profile", &DMWRITE, addObjParentalControlProfile, delObjParentalControlProfile, NULL, browseParentalControlProfileInst, NULL, NULL, tParentalControlProfileObj, tParentalControlProfileParams, NULL, BBFDM_BOTH, NULL},
{0}
};

DMLEAF tParentalControlParams[] = {
/* PARAM, permission, type, getvalue, setvalue, bbfdm_type */
{"Enable", &DMWRITE, DMT_BOOL, get_ParentalControl_Enable, set_ParentalControl_Enable, BBFDM_BOTH},
{"ProfileNumberOfEntries", &DMREAD, DMT_UNINT, get_ParentalControl_ProfileNumberOfEntries, NULL, BBFDM_BOTH},
{"URLBundleNumberOfEntries", &DMREAD, DMT_UNINT, get_ParentalControl_URLBundleNumberOfEntries, NULL, BBFDM_BOTH},
{0}
};

/* *** Device.{BBF_VENDOR_PREFIX}ParentalControl.URLBundle.{i}. *** */
DMLEAF tParentalControlURLBundleParams[] = {
/* PARAM, permission, type, getvalue, setvalue, bbfdm_type */
{"Name", &DMWRITE, DMT_STRING, get_ParentalControlURLBundle_Name, set_ParentalControlURLBundle_Name, BBFDM_BOTH, DM_FLAG_UNIQUE|DM_FLAG_LINKER},
{"DownloadURL", &DMWRITE, DMT_STRING, get_ParentalControlURLBundle_DownloadURL, set_ParentalControlURLBundle_DownloadURL, BBFDM_BOTH},
{"CustomURL", &DMWRITE, DMT_STRING, get_ParentalControlURLBundle_CustomURL, set_ParentalControlURLBundle_CustomURL, BBFDM_BOTH},
{0}
};

/* *** Device.{BBF_VENDOR_PREFIX}ParentalControl.Profile.{i}. *** */
DMOBJ tParentalControlProfileObj[] = {
/* OBJ, permission, addobj, delobj, checkdep, browseinstobj, nextdynamicobj, dynamicleaf, nextobj, leaf, linker, bbfdm_type, uniqueKeys */
{"URLFilter", &DMWRITE, addObjParentalControlProfileURLFilter, delObjParentalControlProfileURLFilter, NULL, browseParentalControlProfileURLFilterInst, NULL, NULL, NULL, tParentalControlProfileURLFilterParams, NULL, BBFDM_BOTH, NULL},
{"BedTime", &DMREAD, NULL, NULL, NULL, NULL, NULL, NULL, tParentalControlProfileBedTimeObj, tParentalControlProfileBedTimeParams, NULL, BBFDM_BOTH, NULL},
{"InternetBreak", &DMREAD, NULL, NULL, NULL, NULL, NULL, NULL, NULL, tParentalControlProfileInternetBreakParams, NULL, BBFDM_BOTH, NULL},
{0}
};

DMLEAF tParentalControlProfileParams[] = {
/* PARAM, permission, type, getvalue, setvalue, bbfdm_type */
{"Alias", &DMWRITE, DMT_STRING, get_ParentalControlProfile_Alias, set_ParentalControlProfile_Alias, BBFDM_BOTH, DM_FLAG_UNIQUE|DM_FLAG_LINKER},
{"Name", &DMWRITE, DMT_STRING, get_ParentalControlProfile_Name, set_ParentalControlProfile_Name, BBFDM_BOTH},
{"IconName", &DMWRITE, DMT_STRING, get_ParentalControlProfile_IconName, set_ParentalControlProfile_IconName, BBFDM_BOTH},
{"HostList", &DMWRITE, DMT_STRING, get_ParentalControlProfile_HostList, set_ParentalControlProfile_HostList, BBFDM_BOTH},
{"URLFilterNumberOfEntries", &DMREAD, DMT_UNINT, get_ParentalControlProfile_URLFilterNumberOfEntries, NULL, BBFDM_BOTH},
{0}
};

/* *** Device.{BBF_VENDOR_PREFIX}ParentalControl.Profile.{i}.URLFilter.{i}. *** */
DMLEAF tParentalControlProfileURLFilterParams[] = {
/* PARAM, permission, type, getvalue, setvalue, bbfdm_type */
{"Enable", &DMWRITE, DMT_BOOL, get_ParentalControlProfileURLFilter_Enable, set_ParentalControlProfileURLFilter_Enable, BBFDM_BOTH},
{"AccessPolicy", &DMWRITE, DMT_STRING, get_ParentalControlProfileURLFilter_AccessPolicy, set_ParentalControlProfileURLFilter_AccessPolicy, BBFDM_BOTH},
{"Order", &DMWRITE, DMT_UNINT, get_ParentalControlProfileURLFilter_Order, set_ParentalControlProfileURLFilter_Order, BBFDM_BOTH},
{"FilterText", &DMWRITE, DMT_STRING, get_ParentalControlProfileURLFilter_FilterText, set_ParentalControlProfileURLFilter_FilterText, BBFDM_BOTH},
{"URLBundleRef", &DMWRITE, DMT_STRING, get_ParentalControlProfileURLFilter_URLBundleRef, set_ParentalControlProfileURLFilter_URLBundleRef, BBFDM_BOTH, DM_FLAG_REFERENCE},
{"ScheduleRef", &DMWRITE, DMT_STRING, get_ParentalControlProfileURLFilter_ScheduleRef, set_ParentalControlProfileURLFilter_ScheduleRef, BBFDM_BOTH, DM_FLAG_REFERENCE},
{0}
};

/* *** Device.{BBF_VENDOR_PREFIX}ParentalControl.Profile.{i}.BedTime. *** */
DMOBJ tParentalControlProfileBedTimeObj[] = {
/* OBJ, permission, addobj, delobj, checkdep, browseinstobj, nextdynamicobj, dynamicleaf, nextobj, leaf, linker, bbfdm_type, uniqueKeys */
{"Schedule", &DMWRITE, addObjParentalControlProfileBedTimeSchedule, delObjParentalControlProfileBedTimeSchedule, NULL, browseParentalControlProfileBedTimeScheduleInst, NULL, NULL, NULL, tParentalControlProfileBedTimeScheduleParams, NULL, BBFDM_BOTH, NULL},
{0}
};

DMLEAF tParentalControlProfileBedTimeParams[] = {
/* PARAM, permission, type, getvalue, setvalue, bbfdm_type */
{"Enable", &DMWRITE, DMT_BOOL, get_ParentalControlProfileBedTime_Enable, set_ParentalControlProfileBedTime_Enable, BBFDM_BOTH},
{"ScheduleNumberOfEntries", &DMREAD, DMT_UNINT, get_ParentalControlProfileBedTime_ScheduleNumberOfEntries, NULL, BBFDM_BOTH},
{0}
};

/* *** Device.{BBF_VENDOR_PREFIX}ParentalControl.Profile.{i}.BedTime.Schedule.{i}. *** */
DMLEAF tParentalControlProfileBedTimeScheduleParams[] = {
/* PARAM, permission, type, getvalue, setvalue, bbfdm_type */
{"Alias", &DMWRITE, DMT_STRING, get_ParentalControlProfileBedTimeSchedule_Alias, set_ParentalControlProfileBedTimeSchedule_Alias, BBFDM_BOTH, DM_FLAG_UNIQUE},
{"Enable", &DMWRITE, DMT_STRING, get_ParentalControlProfileBedTimeSchedule_Enable, set_ParentalControlProfileBedTimeSchedule_Enable, BBFDM_BOTH},
{"Name", &DMWRITE, DMT_STRING, get_ParentalControlProfileBedTimeSchedule_Name, set_ParentalControlProfileBedTimeSchedule_Name, BBFDM_BOTH},
{"Days", &DMWRITE, DMT_STRING, get_ParentalControlProfileBedTimeSchedule_Days, set_ParentalControlProfileBedTimeSchedule_Days, BBFDM_BOTH},
{"StartTime", &DMWRITE, DMT_STRING, get_ParentalControlProfileBedTimeSchedule_StartTime, set_ParentalControlProfileBedTimeSchedule_StartTime, BBFDM_BOTH},
{"EndTime", &DMWRITE, DMT_STRING, get_ParentalControlProfileBedTimeSchedule_EndTime, set_ParentalControlProfileBedTimeSchedule_EndTime, BBFDM_BOTH},
{0}
};

/* *** Device.{BBF_VENDOR_PREFIX}ParentalControl.Profile.{i}.InternetBreak. *** */
DMLEAF tParentalControlProfileInternetBreakParams[] = {
/* PARAM, permission, type, getvalue, setvalue, bbfdm_type */
{"Enable", &DMWRITE, DMT_BOOL, get_ParentalControlProfileInternetBreak_Enable, set_ParentalControlProfileInternetBreak_Enable, BBFDM_BOTH},
{"ScheduleRef", &DMWRITE, DMT_STRING, get_ParentalControlProfileInternetBreak_ScheduleRef, set_ParentalControlProfileInternetBreak_ScheduleRef, BBFDM_BOTH, DM_FLAG_REFERENCE},
{0}
};

