/*
 * utils.c: URL filter misc. utils
 *
 * Copyright (C) 2024 iopsys Software Solutions AB. All rights reserved.
 *
 * Author: Rahul Thakur <rahul.thakur@iopsys.eu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */
#include <arpa/inet.h>
#include <netinet/in.h>
#include <time.h>

#include "utils.h"

/* return 1 if IP string is valid, else return 0 */
int validate_ip(char *ip_str)
{
	struct sockaddr_in sa;

	return inet_pton(AF_INET, ip_str, &(sa.sin_addr));

}

/* return 1 if IPv6 string is valid, else return 0 */
int validate_ipv6(char *ip6_str)
{
	struct sockaddr_in6 sa6;

	return inet_pton(AF_INET6, ip6_str, &(sa6.sin6_addr));
}

// Function to convert wildcard pattern to regex pattern
void wildcard_to_regex(const char *wildcard, char *regex)
{
	char *p = regex;
	*p++ = '^'; // Match the start of the string

	while (*wildcard) {
		if (*wildcard == '*') {
			*p++ = '.';
			*p++ = '*';
		} else if (*wildcard == '.') {
			*p++ = '\\';
			*p++ = '.';
		} else {
			*p++ = *wildcard;
		}
		wildcard++;
	}

	*p++ = '$'; // Match the end of the string
	*p = '\0';

	return;
}

// Helper function to print the details of a schedule
void print_schedule(struct schedule *sched, struct blob_buf *b)
{
	void *schedule_table = blobmsg_open_table(b, "schedule");

	blobmsg_add_string(b, "enable", sched->enable ? "enabled" : "disabled");
	blobmsg_add_u32(b, "duration", sched->duration);
	blobmsg_add_string(b, "start_time", sched->start_time);

	void *array = NULL;
	array = blobmsg_open_array(b, "weekdays");

	size_t j;
	for (j = 0; j < 7; j++) {
		blobmsg_add_string(b, NULL, sched->day[j] == FILTER_ACTIVE ? "active" : "inactive");
	}

	blobmsg_close_array(b, array);
	blobmsg_close_table(b, schedule_table);
}

// Function to print URL filters, bundles, and schedules for a profile
void print_urlfilter(struct urlfilter *filter, struct blob_buf *b)
{
	void *urlfilter_table = blobmsg_open_table(b, "urlfilter");

	blobmsg_add_string(b, "parent", filter->dm_parent);
	blobmsg_add_string(b, "access_policy", filter->access_policy ? "allow" : "deny");

	void *array = NULL;
	array = blobmsg_open_array(b, "filter_text");

	// Print filter text
	struct url_list *filter_text;
	list_for_each_entry(filter_text, &filter->filter_text, list) {
		blobmsg_add_string(b, NULL, filter_text->url);
	}
	blobmsg_close_array(b, array);

	// Print URL bundle
	if (filter->bundle) {
		void *urlbundle_table = blobmsg_open_table(b, "urlbundle");
		blobmsg_add_string(b, "name", filter->bundle->name);

		void *array_inner = blobmsg_open_array(b, "custom_url");
		struct url_list *custom_url;
		list_for_each_entry(custom_url, &(filter->bundle->custom_url), list) {
			blobmsg_add_string(b, NULL, custom_url->url);
		}
		blobmsg_close_array(b, array_inner);

		array_inner = blobmsg_open_array(b, "download_url");
		struct url_list *download_url;
		list_for_each_entry(download_url, &(filter->bundle->download_url), list) {
			blobmsg_add_string(b, NULL, download_url->url);
		}
		blobmsg_close_array(b, array_inner);
		blobmsg_close_table(b, urlbundle_table);
	}

	// Print the schedule list
	struct schedule *schedule = NULL;
	list_for_each_entry(schedule, &filter->schedules, list) {
		print_schedule(schedule, b);
	}

	blobmsg_close_table(b, urlfilter_table);
}

// Function to print profiles, their hosts, filters, and corresponding bundle and schedule information
void print_profile(struct profile *prof, struct blob_buf *b)
{
	void *profile_table = NULL;
	void *host_array = NULL;

	profile_table = blobmsg_open_table(b, "");
	blobmsg_add_string(b, "profile", prof->name);

	host_array = blobmsg_open_array(b, "host");

	// Print associated hosts
	struct host *host;
	list_for_each_entry(host, &prof->hosts, list) {
		blobmsg_add_string(b, NULL, host->host_identifier);
	}
	blobmsg_close_array(b, host_array);

	// Print associated URL filters
	struct urlfilter *filter = NULL;
	list_for_each_entry(filter, &prof->filters, list) {
		print_urlfilter(filter, b);
	}
	blobmsg_close_table(b, profile_table);
}

// Fetches the current day and time
// @param day_index index of the day of week
// @return current time in seconds.
unsigned int get_current_day_and_time(unsigned int *day_index)
{
	const char *weekdays[] = {"Sunday",
				"Monday", "Tuesday",
				"Wednesday", "Thursday",
				"Friday", "Saturday"};
	char day[MAX_LEN_WEEKDAY];
	time_t cur_time;
	struct tm *loc_time;
	unsigned int i = 0;

	// First fetch the day
	memset(day, '\0', MAX_LEN_WEEKDAY);
	cur_time = time(NULL);
	loc_time = localtime(&cur_time);
	strftime(day, MAX_LEN_WEEKDAY, "%A", loc_time);
	day[strcspn(day, "\r\n")] = 0;

	// Fetch index for the day.
	for (i = 0; i < MAX_DAYS; i++) {
		if (strncasecmp(day, weekdays[i], sizeof(day)) == 0) {
			*day_index = i;
			break;
		}
	}

	// Get the current time in seconds format.
	return (unsigned int)(((loc_time->tm_hour * 60) + loc_time->tm_min) * 60 + loc_time->tm_sec);
}

// Function gets the qname ie the hostname for which the dns query
// has been generated, from the packet
// @param data payload
// @domaaain_name variable to store the url
// @return nothing
char *get_domain_name(char data[])
{
	// The content of the QNAME is of the format 3www8mydomain3com0
	// instead of www.mydomain.com. Also this is a variable length
	// field to fit the hostname. So we need to parse the data to
	// get the url for which the dns query has been initiated.
	int i = 0, j = 0, start = 0;
	int toread = data[0];

	i++;

	while (toread != 0) {
		for (; i <= start+toread; i++) {
			data[j] = data[i];
			j++;
		}

		// Get next chunk
		data[j] = '.';
		j++;
		toread = data[i];
		start = i;
		i++;
	}

	data[i - 2] = '\0';
	return data;
}

static void free_global_dns_cache(void)
{
	if (!global_dns_cache) {
		return; // Check if cache itself is NULL
	}

	for (size_t i = 0; i < dns_cache_cnt; i++) {
		// Free each element in the ip_addr array, if it exists
		if (global_dns_cache[i].ip_addr) {
			for (size_t j = 0; j < global_dns_cache[i].ip_addr_cnt; j++) {
				free(global_dns_cache[i].ip_addr[j]);
			}
			free(global_dns_cache[i].ip_addr); // Free the ip_addr array itself
		}

		// Free each element in the ipv6_addr array, if it exists
		if (global_dns_cache[i].ipv6_addr) {
			for (size_t j = 0; j < global_dns_cache[i].ipv6_addr_cnt; j++) {
				free(global_dns_cache[i].ipv6_addr[j]);
			}
			free(global_dns_cache[i].ipv6_addr); // Free the ipv6_addr array itself
		}

		// Free each element in the cname array, if it exists
		if (global_dns_cache[i].cname) {
			for (size_t j = 0; j < global_dns_cache[i].cname_cnt; j++) {
				free(global_dns_cache[i].cname[j]);
			}
			free(global_dns_cache[i].cname); // Free the cname array itself
		}
	}

	free(global_dns_cache); // Free the overall cache array
}

static void free_host_list(struct list_head *host_list)
{
	struct host *host, *tmp;

	// Iterate through the list of hosts and free each one
	list_for_each_entry_safe(host, tmp, host_list, list) {
		// No dynamically allocated members inside 'host', but if needed, free here
		list_del(&host->list);  // Remove from the list
		free(host);  // Free the structure
	}
}

static void free_schedule_list(struct list_head *schedule)
{
	struct schedule *sched = NULL, *tmp = NULL;

	// Iterate through the list of schedules and free each one
	list_for_each_entry_safe(sched, tmp, schedule, list) {
		// Free any additional lists or structures inside 'schedule' if needed
		list_del(&sched->list);  // Remove from the list
		free(sched);  // Free the structure
	}
}

static void free_url_list(struct list_head *url_list)
{
	struct url_list *url_entry = NULL, *tmp = NULL;

	// Iterate through the list of URLs and free each one
	list_for_each_entry_safe(url_entry, tmp, url_list, list) {
		list_del(&url_entry->list);  // Remove from the list
		free(url_entry);  // Free the structure
	}
}

static void free_urlbundle_list(void)
{
	struct urlbundle *bundle, *tmp;

	// Iterate through the list of URL bundles and free each one
	list_for_each_entry_safe(bundle, tmp, &bundles, list) {
		// Free the download_url and custom_url lists
		free_url_list(&bundle->download_url);
		free_url_list(&bundle->custom_url);

		list_del(&bundle->list);  // Remove from the list
		free(bundle);  // Free the structure
	}
}

static void free_urlfilter_list(struct list_head *filter_list)
{
	struct urlfilter *filter = NULL, *tmp = NULL;

	// Iterate through the list of filters and free each one
	list_for_each_entry_safe(filter, tmp, filter_list, list) {
		// Free the list of filter URLs
		free_url_list(&filter->filter_text);
		free_schedule_list(&filter->schedules);

		// Do not free bundles here since they might be shared
		// between multiple filters; free them separately if needed.

		list_del(&filter->list);  // Remove from the list
		free(filter);  // Free the structure
	}
}

static void free_profile_list(void)
{
	struct profile *prof, *tmp;

	// Iterate through the list of profiles and free each one
	list_for_each_entry_safe(prof, tmp, &profiles, list) {
		// Free the list of hosts within the profile
		free_host_list(&prof->hosts);

		// Free the list of filters associated with the profile
		free_urlfilter_list(&prof->filters);

		list_del(&prof->list);  // Remove from the list
		free(prof);  // Free the profile structure
	}
}

void free_all(void)
{
	// Free each list properly
	free_profile_list();
	free_urlbundle_list();
	free_global_dns_cache();
}


